@@ -1,15 +1,16 @@
 import sqlalchemy as sa
 from sqlalchemy_utils.functions import create_database
-from pyg_base import cache, cfg_read, as_list, dictable, lower, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique
+from pyg_base import cache, cfg_read, as_list, dictable, lower, last, loop, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique
 from pyg_encoders import as_reader, as_writer, dumps, loads
 from sqlalchemy import Table, Column, Integer, String, MetaData, Identity, Float, DATE, DATETIME, TIME, select, func, not_, desc, asc
 from sqlalchemy.orm import Session
 from sqlalchemy.types import NUMERIC, FLOAT, INT
 import datetime
 from copy import copy
-from pyg_base import logger
+from pyg_base import logger, is_regex
 from functools import partial
 import pandas as pd
+import pickle
 
 _id = '_id'
 _doc = 'doc'
@@ -29,6 +30,35 @@ _types = {str: String, 'str' : String,
 
 _orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
+def pickle_loads(value):
+    if isinstance(value, dict):
+        return type(value)({k: pickle_loads(v) for k, v in value.items()})
+    elif isinstance(value, (list, tuple)):
+        return type(value)([pickle_loads(v) for v in value])        
+    elif isinstance(value, bytes):
+        return pickle.loads(value)
+    else:
+        return value
+
+def _pair_wise_filter(t, k, v):
+    """
+    t is a sqlalchemy column.c filter
+    """
+    if is_regex(v):
+        p = v.pattern
+        if not p.startswith('%'):
+            p = '%' + p
+        if not p.endswith('%'):
+            p = p + '%'
+        return t[k].like(p)
+    elif is_str(v) and v.startswith('%') and v.endswith('%'):
+        return t[k].like(v)
+    elif isinstance(v, list):
+        return sa.or_(*[_pair_wise_filter(t, k, i) for i in v])
+    else:
+        return t[k] == v
+
+
 @cache
 def _servers():
     res = cfg_read().get('sql_server', {})
@@ -221,11 +251,6 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
     mode : int, optional
         NOT IMPLEMENTED CURRENTLY
 
-    Raises
-    ------
-    ValueError
-        DESCRIPTION.
-
     Returns
     -------
     res : sql_cursor
@@ -234,6 +259,7 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
 
     Example: simple table creation
     ---------
+    >>> from pyg import * 
     >>> table = sql_table(table = 'test', nullable = dict(a = int, b = str), db = 'db')
     >>> table.insert(dict(a = 1, b = 'a'))
     >>> assert len(table) == 1
@@ -244,6 +270,7 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
     >>> table.inc(b = 'a').delete()
     >>> assert len(table) == 2
     >>> assert table.distinct('b') == ['b']
+
     >>> table.drop()
     
     Example: ensure sql_table defaults for a doc store if 'doc' exists in existing table or no data columns are specified on creation
@@ -253,6 +280,8 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
     table = sql_table(table = 'test', db = 'db', schema = 'dbo')
     assert tbl.doc == 'doc'
     table.drop()
+
+    Example: simple join
         
     """
     if isinstance(table, str):
@@ -355,7 +384,7 @@ class sql_cursor(object):
     - maintainance of a table where records are unique per specified primary keys while we auto-archive old data
     - creation of a full no-sql like document-store
 
-    pyg-sql "abandons" the relational part of SQL: we make using a single table extremely easy while forgo any multiple-tables-relations completely.
+    pyg-sql supports simple joins but not much more than that.
     
     ## access simplification
     
@@ -416,8 +445,72 @@ class sql_cursor(object):
     >>> assert len(t.where(t.c.age > 30)) == 2  # can filter using the standard sql-alchemy "where" statement 
 
 
-    ## insertion of "documents" into string columns...
+    Example: simple joining
+    -----------------------
+    >>> from pyg import * 
+    >>> students = sql_table(table = 'students', pk = ['id'], non_null = ['name', 'surname'], db = 'db')
+    >>> students.insert([dict(name = 'ann', surname = 'andrews', id = 'a'),
+                         dict(name = 'ben', surname = 'baxter', id = 'b'),
+                         dict(name = 'charles', surname = 'cohen', id = 'c')])
+
+    >>> subjects = sql_table(table = 'subjects', pk = ['id'], non_null = ['subject', 'teacher'], db = 'db')
+    >>> subjects.insert([dict(subject = 'maths', teacher = 'mandy miles', id = 'm'),
+                         dict(subject = 'geography', teacher = 'george graham', id = 'g'),
+                         dict(subject = 'zoology', teacher = 'zoe zhenya', id = 'z')])
+
+    >>> student_subject = sql_table(table = 'student_subject', non_null = {'student_id':str, 'subject_id':str, 'score': int})
+    >>> student_subject.insert([dict(student_id = 'a', subject_id = 'm', score = 90),
+                                dict(student_id = 'a', subject_id = 'z', score = 85),
+                                dict(student_id = 'b', subject_id = 'm', score = 70),
+                                dict(student_id = 'b', subject_id = 'g', score = 60),
+                                dict(student_id = 'c', subject_id = 'g', score = 50),
+                                dict(student_id = 'c', subject_id = 'z', score = 45),
+                                ])
     
+
+    >>> self = student_subject.join(students, dict(student_id = 'id'))
+
+    self[::]
+    id|name   |score|student_id|subject_id|surname
+    a |ann    |90   |a         |m         |andrews
+    a |ann    |85   |a         |z         |andrews
+    b |ben    |70   |b         |m         |baxter 
+    b |ben    |60   |b         |g         |baxter 
+    c |charles|50   |c         |g         |cohen  
+    c |charles|45   |c         |z         |cohen  
+
+    self.inc(surname = '%co%')[::]
+    dictable[2 x 6]
+    id|name   |score|student_id|subject_id|surname
+    c |charles|50   |c         |g         |cohen  
+    c |charles|45   |c         |z         |cohen  
+
+    >>> self.inc(score = [90,50])[::]
+    dictable[2 x 6]
+    id|name   |score|student_id|subject_id|surname
+    a |ann    |90   |a         |m         |andrews
+    c |charles|50   |c         |g         |cohen  
+
+    ## double join
+    >>> student_subject.join(students, dict(student_id = 'id')).join(subjects, dict(subject_id = 'id'))[::]
+    id|id_1|name   |score|student_id|subject  |subject_id|surname|teacher      
+    a |m   |ann    |90   |a         |maths    |m         |andrews|mandy miles  
+    a |z   |ann    |85   |a         |zoology  |z         |andrews|zoe zhenya   
+    b |m   |ben    |70   |b         |maths    |m         |baxter |mandy miles  
+    b |g   |ben    |60   |b         |geography|g         |baxter |george graham
+    c |g   |charles|50   |c         |geography|g         |cohen  |george graham
+    c |z   |charles|45   |c         |zoology  |z         |cohen  |zoe zhenya       
+
+    >>> full_join = student_subject.join(students, dict(student_id = 'id')).join(subjects, dict(subject_id = 'id'))
+    >>> full_join.inc(TEACHER = '%mandy%', SCORE = 90)
+    sql_cursor: db.dbo.student_subject  
+    SELECT student_id, subject_id, score, dbo.students.id, dbo.students.name, dbo.students.surname, dbo.subjects.id AS id_1, dbo.subjects.subject, dbo.subjects.teacher 
+    FROM dbo.student_subject JOIN dbo.students ON student_id = dbo.students.id JOIN dbo.subjects ON subject_id = dbo.subjects.id 
+    WHERE dbo.subjects.teacher LIKE "%mandy%" AND score = 90
+    1 records
+
+    :Example: insertion of "documents" into string columns...    
+    ----------------------------------------------------------
     It is important to realise that we already have much flexibility behind the scene in using "documents" inside string columns:
 
     >>> t = t.delete()
@@ -557,7 +650,9 @@ class sql_cursor(object):
     >>> assert list(read_from_db.salary.values) == [100, 200, 300]
     >>> assert list(read_from_file.values) == [100, 200, 300]
     """
-    def __init__(self, table, schema = None, db = None, engine = None, server = None, spec = None, selection = None, order = None, reader = None, writer = None, pk = None, doc = None, **_):
+    def __init__(self, table, schema = None, db = None, engine = None, server = None, 
+                 spec = None, selection = None, order = None, joint = None, reader = None, writer = None, 
+                 pk = None, doc = None, **_):
         """
         Parameters
         ----------
@@ -596,6 +691,7 @@ class sql_cursor(object):
             selection = table.selection if selection is None else selection
             schema = table.schema if schema is None else schema
             order = table.order if order is None else order
+            joint = table.joint if joint is None else joint
             reader = table.reader if reader is None else reader
             writer = table.writer if writer is None else writer
             pk = table.pk if pk is None else pk
@@ -605,11 +701,12 @@ class sql_cursor(object):
         self.table = table
         self.schema = schema
         self.db = db
-        self.server = get_server(server)
+        self.server = server
         self.engine = engine or get_engine(db = self.db, server = self.server)
         self.spec = spec
         self.selection = selection
         self.order = order
+        self.joint = joint
         self.reader = reader
         self.writer = writer
         self.pk = pk
@@ -663,8 +760,43 @@ class sql_cursor(object):
         """
         ids = self._ids
         return sorted([c.name for c in self.tbl.columns if c.nullable is False and c.name not in ids])
-        
-    
+
+
+    @property
+    def _columns(self):
+        cols = self.columns
+        return dict(zip(lower(cols), cols))
+
+    def _joint_columns(self):
+        """
+        returns a mapping from lower col to both column name and table
+
+        """
+        rs = dictable(table = [j[0] for j in self.joint])
+        res = rs(grp = lambda table: dictable(column = table.columns, col = lower(table.columns))).ungroup().listby('col')
+        duplicates = dict(res.inc(lambda column: len(column)>1)['col','table'])
+        res = res.inc(lambda column: len(column)==1).do(last, 'column', 'table')
+        jcol = dict(res['col', 'column'])
+        jtbl = dict(res['col', 'table'])
+        return jcol, jtbl, duplicates
+
+    def _kw(self, **kwargs):
+        columns = self._columns
+        jcol, jtbl, duplicates = self._joint_columns()
+        conditions = []
+        for k, v in kwargs.items():
+            k = k.lower()
+            if k in columns:
+                conditions.append(_pair_wise_filter(self.table.c, columns[k], v))
+            elif k in duplicates:
+                raise ValueError(f'column {k} exists in multiple tables: \n{duplicates[k]}')
+            elif k in jcol:
+                conditions.append(_pair_wise_filter(jtbl[k].c, jcol[k], v))
+            else:
+                raise ValueError(f'column {k} not found')
+        return sa.and_(*conditions)
+
+
     def _c(self, expression):
         """
         converts an expression to a sqlalchemy filtering expression
@@ -673,11 +805,13 @@ class sql_cursor(object):
         ---------
         >>> expression = dict(a = 1, b = 2)
         >>> assert t._c(expression) == sa.and_(t.c.a == 1, t.c.b == 2)
+        c = table.insert_many(dictable(a = ['a','aa','b', 'c'], b = ['bb','b1','c','d']))
         """
         if isinstance(expression, dict):
-            expression = self._col(expression)
-            t = self.table.c    
-            return sa.and_(*[sa.or_(*[t[k] == i for i in v]) if isinstance(v, list) else t[k] == self._c(v) for k,v in expression.items()]) 
+            return self._kw(**expression)
+            # expression = self._col(expression)
+            # t = self.table.c
+            # return sa.and_(*[_pair_wise_filter(t, k, v) for k,v in expression.items()]) 
         elif isinstance(expression, (list, tuple)):
             return sa.or_(*[self._c(v) for v in expression])            
         else:
@@ -785,10 +919,14 @@ class sql_cursor(object):
         if self.selection is None:
             return self.select(self.columns - other)
         elif is_strs(self.selection):
-            return self.select(ulist(as_list(self.selection)) - other)
+            selection = self._col(as_list(self.selection))
+            return self.select(ulist(selection) - other)
         else:
-            raise ValueError('cannot subtract these columns while the selection is non empty, use self.select() to reset selection')
-    
+            selection = as_list(self.selection)
+            other = as_list(other)
+            selection = [col for col in selection if col.name not in other]
+            return self.select(selection)
+            
     def drop(self):
         logger.info('dropping table: %s.%s.%s'%(_database(self.db), 
                                                   self.schema, 
@@ -797,6 +935,29 @@ class sql_cursor(object):
         self.table.drop(bind = self.engine)
 
     
+    def join(self, right, onclose = None, isouter = False, full = False):
+        res = self.copy()
+        res.joint = as_list(self.joint) + [(right, onclose, isouter, full)]
+        return res
+
+    @property
+    def _table(self):
+        """
+        a table supporting join objects
+        """
+        res = self.table
+        for j in as_list(self.joint):            
+            right, onclose, isouter, full = j
+            if isinstance(right, sql_cursor):
+                right = right._table
+            if is_strs(onclose):
+                onclose = as_list(onclose)
+                onclose = sa.and_(*[self.c[col] == right.c[col] for col in onclose])
+            elif is_dict(onclose):
+                onclose = sa.and_(*[self.c[key] == right.c[value] for key, value in onclose.items()])
+            res = res.join(right, onclose, isouter = isouter, full = full)
+        return res
+
     def exc(self, *args, **kwargs):
         """
         Exclude: This returns a table with additional filtering OPPOSITE TO inc. note that you can build it iteratively
@@ -840,7 +1001,7 @@ class sql_cursor(object):
     
     
     def __len__(self):
-        statement = select(func.count()).select_from(self.table)
+        statement = select(func.count()).select_from(self._table)
         if self.spec is not None:
             statement = statement.where(self.spec)
         return list(self.engine.connect().execute(statement))[0][0]
@@ -854,9 +1015,9 @@ class sql_cursor(object):
         if is_str(column):
             return cols[column.lower()]
         elif isinstance(column, (list, tuple)):
-            return [cols[c.lower()] for c in column]
+            return [cols[c.lower()] if isinstance(c,str) else c for c in column]
         elif isinstance(column, dict):
-            return type(column)({cols[k.lower()] : v for k, v in column.items()})
+            return type(column)({cols[k.lower()]: v for k, v in column.items()})
         else:
             raise ValueError('column %s cannot be converted to table columns'%column)
 
@@ -868,7 +1029,7 @@ class sql_cursor(object):
         if len(value) == 0:
             return self
         res = self.copy()
-        res.selection = self._col(as_list(value))
+        res.selection = as_list(value)
         return res
     
     def _enrich(self, doc, columns = None):
@@ -917,8 +1078,8 @@ class sql_cursor(object):
             ids = self._ids
             res_no_ids = type(res)({k : v for k, v in res.items() if k not in ids}) if ids else res
             tbl = self.inc().inc(**doc_id)
-            rows = tbl.sort(ids)._read_statement() ## row format
-            docs = tbl._rows_to_docs(rows, reader = False, load = False) ## do not transform the document, keep in raw format
+            read = tbl.sort(ids)._read_statement() ## row format
+            docs = tbl._rows_to_docs(reader = False, load = False, **read) ## do not transform the document, keep in raw format
             if len(docs) == 0:
                 with self.engine.connect() as conn: 
                     conn.execute(self.table.insert(),[res_no_ids])
@@ -1050,11 +1211,12 @@ class sql_cursor(object):
             stop = stop if stop is None else stop - start
         if stop is not None:
             statement = statement.limit(1+stop)
-        res = list(self.engine.connect().execute(statement))
+        executed = self.engine.connect().execute(statement)
+        columns = list(executed.keys())
+        data = list(executed)
         if start is not None or stop is not None or step is not None:
-            return res[slice(start, stop, step)]
-        else:
-            return res
+            data = data[slice(start, stop, step)]
+        return dict(data = data, columns = columns)
     
     def df(self, decimal2float = True):
         """
@@ -1094,8 +1256,9 @@ class sql_cursor(object):
         """
         statement = self.statement()
         res = self.engine.connect().execute(statement)
+        columns = list(res.keys())
         if _pd_is_old:
-            res = pd.DataFrame(list(res), columns = as_list(self.selection) or self.columns)
+            res = pd.DataFrame(list(res), columns = columns)
         else:
             res = pd.DataFrame(res)
         if decimal2float:
@@ -1106,30 +1269,37 @@ class sql_cursor(object):
         return res
             
     
-    def _rows_to_docs(self, rows, reader = None, load = True):
+    def _rows_to_docs(self, data, reader = None, load = True, columns = None):
         """
         starts at raw values from the database and returns a list of read-dicts (or a single dict) from the database
         """
-        columns = as_list(self.selection) if self.selection else self.columns
+        if columns is None:
+            columns = as_list(self.selection) if self.selection else self.columns
         reader = self._reader(reader)
-        if isinstance(rows, list):
-            return [self._read_row(row, reader = reader, columns = columns, load = load) for row in rows]
+        if isinstance(data, list):
+            return [self._read_row(row, reader = reader, columns = columns, load = load) for row in data]
         else:        
-            return self._read_row(rows, reader = reader, columns = columns, load = load)
+            return self._read_row(data, reader = reader, columns = columns, load = load)
 
     def docs(self, start = None, stop = None, step = None, reader = None):
-        rows = self._read_statement(start = start, stop = stop, step = step)
-        docs = self._rows_to_docs(rows, reader = reader)
+        read = self._read_statement(start = start, stop = stop, step = step)
+        docs = self._rows_to_docs(reader = reader, **read)
         return dictable(docs)
     
     def read(self, item = 0, reader = None):
         value = len(self) + item if item < 0 else item
         statement = self.statement()
         if self.order is None:
-            row = list(self.engine.connect().execute(statement.limit(value+1)))[value]
+            executed = self.engine.connect().execute(statement.limit(value+1))
+            columns = list(executed.keys())
+            data = list(executed)
+            row = data[value]
         else:
-            row = list(self.engine.connect().execute(statement.offset(value).limit(1)))[0]
-        doc = self._rows_to_docs(row, reader = reader)
+            executed = self.engine.connect().execute(statement.offset(value).limit(1))
+            columns = list(executed.keys())
+            data = list(executed)
+            row = data[0]
+        doc = self._rows_to_docs(data = row, reader = reader, columns = columns)
         rtn = self._undock(doc)
         return rtn
 
@@ -1167,9 +1337,9 @@ class sql_cursor(object):
                 reader = False
             else:
                 reader = None
-            rows = self._read_statement(start = start, stop = stop, step = step)
-            docs = self._rows_to_docs(rows, reader = reader)
-            columns = self.columns
+            read = self._read_statement(start = start, stop = stop, step = step)
+            docs = self._rows_to_docs(reader = reader, **read)
+            columns = read['columns'] #self.columns
             res = dictable([self._undock(doc, columns = columns) for doc in docs])
             return res
 
@@ -1291,14 +1461,16 @@ class sql_cursor(object):
         """
         performs a selection based on self.selection
         """
+        table = self._table
         if self.selection is None:
-            statement = select(self.table)
+            statement = select(table)
         elif is_strs(self.selection):               
             c = self.table.c
-            selection = [c[v] for v in as_list(self.selection)]
-            statement = select(selection).select_from(self.table)
+            selection = self._col(as_list(self.selection))
+            selection = [c[v] for v in selection]
+            statement = select(selection).select_from(table)
         else: ## user provided sql alchemy selection object
-            statement = select(self.selection).select_from(self.table)
+            statement = select(self.selection).select_from(table)
         return statement
     
     def statement(self):
@@ -1351,8 +1523,8 @@ class sql_cursor(object):
         ids = self._ids
         if len(res):
             if self._pk and not self._is_deleted(): ## we first copy the existing data out to deleted db
-                rows = self._read_statement() 
-                docs = self._rows_to_docs(rows, reader = False, load = False)
+                read = self._read_statement() 
+                docs = self._rows_to_docs(reader = False, load = False, **read)
                 deleted = datetime.datetime.now()
                 for doc in docs:
                     doc[_deleted] = deleted
@@ -1378,16 +1550,19 @@ class sql_cursor(object):
         """
         return self.table.name
     
+    
     def distinct(self, *keys):
         """
         select DISTINCT *keys FROM TABLE
         """
+        keys = as_list(keys)
         if len(keys) == 0 and self.selection is not None:
             keys = as_list(self.selection)
-        keys = self._col(keys)
+        if is_strs(keys):
+            keys = self._col(keys)
+            keys = [self.table.columns[k] for k in keys]
         session = Session(self.engine)
-        cols = [self.table.columns[k] for k in keys]
-        query = session.query(*cols)
+        query = session.query(*keys)
         if self.spec is not None:
             query = query.where(self.spec)        
         res = query.distinct().all()
@@ -1399,7 +1574,7 @@ class sql_cursor(object):
         statement = self.statement()
         params = statement.compile().params
         prefix = '%s.%s'%(self.schema, self.table.name) if self.schema else self.table.name
-        text = str(statement).replace(prefix + '.','')
+        text = str(statement).replace('"','').replace(prefix + '.','')
         for k,v in params.items():
             text = text.replace(':'+k, '"%s"'%v if is_str(v) else dt2str(v) if is_date(v) else str(v))
             
